/*
	vmsvga.c	screen driver
        video board / chip dependent processing: VMware SVGA II

	Copyright 2015 by SASANO Takayoshi
	This software is distributed under the T-License 2.0.
*/
#include "screen.h"
#include <device/videomode.h>
#include <device/pci.h>
#include <tk/syslib.h>

#define	VIDEOMODE	DM2048x32
#define	SUPPORT_MODEMAP	(1 << VIDEOMODE)

LOCAL	UH	IOaddr;

#define	VENDOR_VMWARE	0x15ad
#define	DEVICE_SVGA2	0x0405

#define	regENABLE	1
#define	regWIDTH	2
#define	regHEIGHT	3
#define	regBPP		7
#define	regPITCH	12
#define	regVRAMSIZE	15
#define	regCONFIG	20
#define	regIRQMASK	33

Inline	void	WriteSVGA(UW index, UW value)
{
	out_w(IOaddr + 0, index);
	out_w(IOaddr + 1, value);
	return;
}

Inline	UW	ReadSVGA(UW index)
{
	out_w(IOaddr + 0, index);
	return in_w(IOaddr + 1);
}

LOCAL	ER	VMSVGAinit(void)
{
	ER	er;

	/* find VMware SVGA II device */
	Vinf.pciaddr = searchPciDev(VENDOR_VMWARE, DEVICE_SVGA2);
	if (Vinf.pciaddr < 0) {
		er = E_NOMDA;
		goto fin0;
	}

	/* enable FrameBuffer and I/O register */
	outPciConfH(Vinf.pciaddr, PCR_COMMAND,
		    inPciConfH(Vinf.pciaddr, PCR_COMMAND) | 0x0003);

	/* gather addresses, FrameBuffer size */
	IOaddr = inPciConfW(Vinf.pciaddr, PCR_BASEADDR_0) & ~0x03;
	Vinf.framebuf_addr =
		(void *)(inPciConfW(Vinf.pciaddr, PCR_BASEADDR_1) & ~0x0f);
	Vinf.framebuf_total = ReadSVGA(regVRAMSIZE);

	/* disable interrupt */
	WriteSVGA(regIRQMASK, 0);

	/* map FrameBuffer */
	er = mapFrameBuf(Vinf.framebuf_addr, Vinf.framebuf_total,
			 &Vinf.f_addr);
	if (er < E_OK) goto fin0;

	er = E_OK;
fin0:
	return er;
}

/* set color map (not supported) */
LOCAL	void	VMSVGAsetcmap(COLOR *cmap, W index, W entries)
{
	/* no support, do nothing */
	return;
}

/* set display mode */
LOCAL	void	VMSVGAsetmode(W flg)
{
	/* exit VMware SVGA II mode, required for warm reboot */
	// XXX the last contents of VGA mode is redisplayed when exiting.
	if (flg < 0) {
		WriteSVGA(regENABLE, 0);
		return;
	}

	/* enter VMware SVGA II mode */
	WriteSVGA(regWIDTH, Vinf.act_width);
	WriteSVGA(regHEIGHT, Vinf.act_height);
	WriteSVGA(regBPP, Vinf.pixbits >> 8);
	WriteSVGA(regENABLE, 1);
	WriteSVGA(regCONFIG, 0);

	/* fix display mode information */
	Vinf.width = Vinf.fb_width = Vinf.act_width;
	Vinf.height = Vinf.fb_height = Vinf.act_height;
	Vinf.rowbytes = Vinf.framebuf_rowb = ReadSVGA(regPITCH);
	Vinf.vramsz = Vinf.framebuf_rowb * Vinf.fb_height;

	return;
}

/* suspend / resume processing */
LOCAL	void	VMSVGAsuspend(BOOL suspend)
{
        /* in the case of suspend, clear Video-RAM content */
	if (suspend) {
		memset(Vinf.f_addr, 0, Vinf.framebuf_total);
	}

	return;
}

/* initialization processing */
EXPORT	W	VMSVGAInit(void)
{
	ER	er;

	/* check device */
	er = VMSVGAinit();
	if (er < E_OK) goto fin0;

	/* set Vinf */
	Vinf.reqmode = Vinf.curmode = VIDEOMODE;
	// Vinf.framebuf_addr is already set
	// Vinf.f_addr is already set
	strncpy(Vinf.chipinf, "VMware SVGA II", L_CHIPINF);
	// Vinf.framebuf_total is already set
	Vinf.attr |= (LINEAR_FRAMEBUF | NEED_FINPROC | NEED_SUSRESPROC);
	Vinf.attr &= ~BPP_24;
	Vinf.fn_setcmap = VMSVGAsetcmap;
	Vinf.fn_setmode = VMSVGAsetmode;
	Vinf.fn_susres = VMSVGAsuspend;
	Vinf.modemap = SUPPORT_MODEMAP;

	/* these values are temporally, fix them at VMSVGAsetmode() */
	Vinf.fb_width = Vinf.framebuf_rowb = VideoHsize(Vinf.reqmode);
	Vinf.fb_height = VideoVsize(Vinf.reqmode);

	er = 1;
	goto fin0;

fin0:
	return er;
}
